# This file is part of QuTiP: Quantum Toolbox in Python.
#
#    Copyright (c) 2011 and later, Paul D. Nation and Robert J. Johansson.
#    All rights reserved.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions are
#    met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of the QuTiP: Quantum Toolbox in Python nor the names
#       of its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
###############################################################################

__all__ = ['wigner', 'qfunc', 'QFunc', 'spin_q_function',
           'spin_wigner', 'wigner_transform']

import numpy as np
import warnings
from numpy import (
    zeros, array, arange, exp, real, conj, pi, copy, sqrt, meshgrid, cos, sin,
)
import scipy.sparse as sp
import scipy.fftpack as ft
import scipy.linalg as la
import scipy.special
from scipy.special import genlaguerre, binom, sph_harm, factorial

import qutip
from qutip import Qobj, ket2dm, jmat
from qutip.parallel import parfor
from qutip.cy.sparse_utils import _csr_get_diag
from qutip.sparse import eigh


def wigner_transform(psi, j, fullparity, steps, slicearray):
    """takes the density matrix or state vector of any finite state and
    generates the Wigner function for that state on a sphere, generating a spin
    Wigner function useful for displaying the quasi-probability for a qubit or
    any qudit. For the standard, continuous-variable Wigner function for
    position and momentum variables, wigner() should be used.

    Parameters
    ----------
        psi : qobj
              a state vector or density matrix.
        j : int
            the total angular momentum of the quantum state.
        fullparity : bool
                     should the parity of the full SU space be used?
        steps : int
                number of points at which the Wigner transform is calculated.
        slicearray : list of str
                     the angle slice to be used for each particle in case of a
                     multi-particle quantum state. 'l' yields an equal angle
                     slice. 'x', 'y' and 'z' angle slices can also be chosen.

    Returns
    ----------
        wigner : list of float
                 the wigner transformation at `steps` different theta and phi.

    Raises
    ------
    ComplexWarning
        This can be ignored as it is caused due to rounding errors.

    Notes
    ------
    See example notebook wigner_visualisation.

    References
    ------
    [1] T. Tilma, M. J. Everitt, J. H. Samson, W. J. Munro,
        and K. Nemoto, Phys. Rev. Lett. 117, 180401 (2016).
    [2] R. P. Rundle, P. W. Mills, T. Tilma, J. H. Samson, and
        M. J. Everitt, Phys. Rev. A 96, 022117 (2017).
    """
    if not (psi.type == 'ket' or psi.type == 'operator' or psi.type == 'bra'):
        raise TypeError('Input state is not a valid operator.')

    if psi.type == 'ket' or psi.type == 'bra':
        rho = ket2dm(psi)
    else:
        rho = psi

    sun = 2   # The order of the SU group

    # calculate total number of particles in quantum state:
    N = np.int32(np.log(np.shape(rho)[0]) / np.log(2 * j + 1))

    theta = np.zeros((N, steps))
    phi = np.zeros((N, steps))

    for i in range(N):
        theta[i, :] = np.linspace(0, np.pi, steps)
        phi[i, :] = np.linspace(0, 2 * np.pi, steps)

    theta, phi = _angle_slice(np.array(slicearray, dtype=str), theta, phi)

    wigner = np.zeros((steps, steps))
    if fullparity:
        pari = _parity(sun**N, j)
    else:
        pari = _parity(sun, j)
    for t in range(steps):
        for p in range(steps):
            wigner[t, p] = np.real(np.trace(rho.data @ _kernelsu2(
                theta[:, t], phi[:, p], N, j, pari, fullparity)))
    return wigner


def _parity(N, j):
    """Private function to calculate the parity of the quantum system.
    """
    if j == 0.5:
        pi = np.identity(N) - np.sqrt((N - 1) * N * (N + 1) / 2) * _lambda_f(N)
        return pi / N
    elif j > 0.5:
        mult = np.int32(2 * j + 1)
        matrix = np.zeros((mult, mult))
        foo = np.ones(mult)
        for n in np.arange(-j, j + 1, 1):
            for l in np.arange(0, mult, 1):
                foo[l] = (2 * l + 1) * qutip.clebsch(j, l, j, n, 0, n)
            matrix[np.int32(n + j), np.int32(n + j)] = np.sum(foo)
        return matrix / mult


def _lambda_f(N):
    """Private function needed for the calculation of the parity.
    """
    matrix = np.sqrt(2 / (N * (N - 1))) * np.identity(N)
    matrix[-1, -1] = - np.sqrt(2 * (N - 1) / N)
    return matrix


def _kernelsu2(theta, phi, N, j, parity, fullparity):
    """Private function that calculates the kernel for the SU2 unitary group.
    """
    U = np.ones(1)
    # calculate the total rotation matrix (tensor product for each particle):
    for i in range(0, N):
        U = np.kron(U, _rotation_matrix(theta[i], phi[i], j))
    if not fullparity:
        op_parity = parity   # The parity for a one particle system
        for i in range(1, N):
            parity = np.kron(parity, op_parity)
    matrix = U @ parity @ U.conj().T
    return matrix


def _rotation_matrix(theta, phi, j):
    """Private function to calculate the rotation operator for the SU2 kernel.
    """
    return la.expm(1j * phi * jmat(j, 'z').full()) @ \
           la.expm(1j * theta * jmat(j, 'y').full())


def _angle_slice(slicearray, theta, phi):
    """Private function to modify theta and phi for angle slicing.
    """
    xind = np.where(slicearray == 'x')
    theta[xind, :] = np.pi - theta[xind, :]
    phi[xind, :] = -phi[xind, :]
    yind = np.where(slicearray == 'y')
    theta[yind, :] = np.pi - theta[yind, :]
    phi[yind, :] = np.pi - phi[yind, :]
    zind = np.where(slicearray == 'z')
    phi[zind, :] = phi[zind, :] + np.pi
    return theta, phi


def wigner(psi, xvec, yvec, method='clenshaw', g=sqrt(2),
           sparse=False, parfor=False):
    """Wigner function for a state vector or density matrix at points
    `xvec + i * yvec`.

    Parameters
    ----------

    state : qobj
        A state vector or density matrix.

    xvec : array_like
        x-coordinates at which to calculate the Wigner function.

    yvec : array_like
        y-coordinates at which to calculate the Wigner function.  Does not
        apply to the 'fft' method.

    g : float
        Scaling factor for `a = 0.5 * g * (x + iy)`, default `g = sqrt(2)`.
        The value of `g` is related to the value of `hbar` in the commutation
        relation `[x, y] = i * hbar` via `hbar=2/g^2` giving the default
        value `hbar=1`.

    method : string {'clenshaw', 'iterative', 'laguerre', 'fft'}
        Select method 'clenshaw' 'iterative', 'laguerre', or 'fft', where 'clenshaw'
        and 'iterative' use an iterative method to evaluate the Wigner functions for density
        matrices :math:`|m><n|`, while 'laguerre' uses the Laguerre polynomials
        in scipy for the same task. The 'fft' method evaluates the Fourier
        transform of the density matrix. The 'iterative' method is default, and
        in general recommended, but the 'laguerre' method is more efficient for
        very sparse density matrices (e.g., superpositions of Fock states in a
        large Hilbert space). The 'clenshaw' method is the preferred method for
        dealing with density matrices that have a large number of excitations
        (>~50). 'clenshaw' is a fast and numerically stable method.

    sparse : bool {False, True}
        Tells the default solver whether or not to keep the input density
        matrix in sparse format.  As the dimensions of the density matrix
        grow, setthing this flag can result in increased performance.

    parfor : bool {False, True}
        Flag for calculating the Laguerre polynomial based Wigner function
        method='laguerre' in parallel using the parfor function.


    Returns
    -------

    W : array
        Values representing the Wigner function calculated over the specified
        range [xvec,yvec].

    yvex : array
        FFT ONLY. Returns the y-coordinate values calculated via the Fourier
        transform.

    Notes
    -----
    The 'fft' method accepts only an xvec input for the x-coordinate.
    The y-coordinates are calculated internally.

    References
    ----------

    Ulf Leonhardt,
    Measuring the Quantum State of Light, (Cambridge University Press, 1997)

    """

    if not (psi.type == 'ket' or psi.type == 'oper' or psi.type == 'bra'):
        raise TypeError('Input state is not a valid operator.')

    if method == 'fft':
        return _wigner_fourier(psi, xvec, g)

    if psi.type == 'ket' or psi.type == 'bra':
        rho = ket2dm(psi)
    else:
        rho = psi

    if method == 'iterative':
        return _wigner_iterative(rho, xvec, yvec, g)

    elif method == 'laguerre':
        return _wigner_laguerre(rho, xvec, yvec, g, parfor)

    elif method == 'clenshaw':
        return _wigner_clenshaw(rho, xvec, yvec, g, sparse=sparse)

    else:
        raise TypeError(
            "method must be either 'iterative', 'laguerre', or 'fft'.")


def _wigner_iterative(rho, xvec, yvec, g=sqrt(2)):
    r"""
    Using an iterative method to evaluate the wigner functions for the Fock
    state :math:`|m><n|`.

    The Wigner function is calculated as
    :math:`W = \sum_{mn} \rho_{mn} W_{mn}` where :math:`W_{mn}` is the Wigner
    function for the density matrix :math:`|m><n|`.

    In this implementation, for each row m, Wlist contains the Wigner functions
    Wlist = [0, ..., W_mm, ..., W_mN]. As soon as one W_mn Wigner function is
    calculated, the corresponding contribution is added to the total Wigner
    function, weighted by the corresponding element in the density matrix
    :math:`rho_{mn}`.
    """

    M = np.prod(rho.shape[0])
    X, Y = meshgrid(xvec, yvec)
    A = 0.5 * g * (X + 1.0j * Y)

    Wlist = array([zeros(np.shape(A), dtype=complex) for k in range(M)])
    Wlist[0] = exp(-2.0 * abs(A) ** 2) / pi

    W = real(rho[0, 0]) * real(Wlist[0])
    for n in range(1, M):
        Wlist[n] = (2.0 * A * Wlist[n - 1]) / sqrt(n)
        W += 2 * real(rho[0, n] * Wlist[n])

    for m in range(1, M):
        temp = copy(Wlist[m])
        Wlist[m] = (2 * conj(A) * temp - sqrt(m) * Wlist[m - 1]) / sqrt(m)

        # Wlist[m] = Wigner function for |m><m|
        W += real(rho[m, m] * Wlist[m])

        for n in range(m + 1, M):
            temp2 = (2 * A * Wlist[n - 1] - sqrt(m) * temp) / sqrt(n)
            temp = copy(Wlist[n])
            Wlist[n] = temp2

            # Wlist[n] = Wigner function for |m><n|
            W += 2 * real(rho[m, n] * Wlist[n])

    return 0.5 * W * g ** 2


def _wigner_laguerre(rho, xvec, yvec, g, parallel):
    r"""
    Using Laguerre polynomials from scipy to evaluate the Wigner function for
    the density matrices :math:`|m><n|`, :math:`W_{mn}`. The total Wigner
    function is calculated as :math:`W = \sum_{mn} \rho_{mn} W_{mn}`.
    """

    M = np.prod(rho.shape[0])
    X, Y = meshgrid(xvec, yvec)
    A = 0.5 * g * (X + 1.0j * Y)
    W = zeros(np.shape(A))

    # compute wigner functions for density matrices |m><n| and
    # weight by all the elements in the density matrix
    B = 4 * abs(A) ** 2
    if sp.isspmatrix_csr(rho.data):
        # for compress sparse row matrices
        if parallel:
            iterator = (
                (m, rho, A, B) for m in range(len(rho.data.indptr) - 1))
            W1_out = parfor(_par_wig_eval, iterator)
            W += sum(W1_out)
        else:
            for m in range(len(rho.data.indptr) - 1):
                for jj in range(rho.data.indptr[m], rho.data.indptr[m + 1]):
                    n = rho.data.indices[jj]

                    if m == n:
                        W += real(rho[m, m] * (-1) ** m * genlaguerre(m, 0)(B))

                    elif n > m:
                        W += 2.0 * real(rho[m, n] * (-1) ** m *
                                        (2 * A) ** (n - m) *
                                        sqrt(factorial(m) / factorial(n)) *
                                        genlaguerre(m, n - m)(B))
    else:
        # for dense density matrices
        B = 4 * abs(A) ** 2
        for m in range(M):
            if abs(rho[m, m]) > 0.0:
                W += real(rho[m, m] * (-1) ** m * genlaguerre(m, 0)(B))
            for n in range(m + 1, M):
                if abs(rho[m, n]) > 0.0:
                    W += 2.0 * real(rho[m, n] * (-1) ** m *
                                    (2 * A) ** (n - m) *
                                    sqrt(factorial(m) / factorial(n)) *
                                    genlaguerre(m, n - m)(B))

    return 0.5 * W * g ** 2 * np.exp(-B / 2) / pi


def _par_wig_eval(args):
    """
    Private function for calculating terms of Laguerre Wigner function
    using parfor.
    """
    m, rho, A, B = args
    W1 = zeros(np.shape(A))
    for jj in range(rho.data.indptr[m], rho.data.indptr[m + 1]):
        n = rho.data.indices[jj]

        if m == n:
            W1 += real(rho[m, m] * (-1) ** m * genlaguerre(m, 0)(B))

        elif n > m:
            W1 += 2.0 * real(rho[m, n] * (-1) ** m *
                             (2 * A) ** (n - m) *
                             sqrt(factorial(m) / factorial(n)) *
                             genlaguerre(m, n - m)(B))
    return W1


def _wigner_fourier(psi, xvec, g=np.sqrt(2)):
    """
    Evaluate the Wigner function via the Fourier transform.
    """
    if psi.type == 'bra':
        psi = psi.dag()
    if psi.type == 'ket':
        return _psi_wigner_fft(psi.full(), xvec, g)
    elif psi.type == 'oper':
        eig_vals, eig_vecs = eigh(psi.full())
        W = 0
        for ii in range(psi.shape[0]):
            W1, yvec = _psi_wigner_fft(
                np.reshape(eig_vecs[:, ii], (psi.shape[0], 1)), xvec, g)
            W += eig_vals[ii] * W1
        return W, yvec


def _psi_wigner_fft(psi, xvec, g=sqrt(2)):
    """
    FFT method for a single state vector.  Called multiple times when the
    input is a density matrix.
    """
    n = len(psi)
    A = _osc_eigen(n, xvec * g / np.sqrt(2))
    xpsi = np.dot(psi.T, A)
    W, yvec = _wigner_fft(xpsi, xvec * g / np.sqrt(2))
    return (0.5 * g ** 2) * np.real(W.T), yvec * np.sqrt(2) / g


def _wigner_fft(psi, xvec):
    """
    Evaluates the Fourier transformation of a given state vector.
    Returns the corresponding density matrix and range
    """
    n = 2*len(psi.T)
    r1 = np.concatenate((np.array([[0]]),
                        np.fliplr(psi.conj()),
                        np.zeros((1, n//2 - 1))), axis=1)
    r2 = np.concatenate((np.array([[0]]), psi,
                        np.zeros((1, n//2 - 1))), axis=1)
    w = la.toeplitz(np.zeros((n//2, 1)), r1) * \
        np.flipud(la.toeplitz(np.zeros((n//2, 1)), r2))
    w = np.concatenate((w[:, n//2:n], w[:, 0:n//2]), axis=1)
    w = ft.fft(w)
    w = np.real(np.concatenate((w[:, 3*n//4:n+1], w[:, 0:n//4]), axis=1))
    p = np.arange(-n/4, n/4)*np.pi / (n*(xvec[1] - xvec[0]))
    w = w / (p[1] - p[0]) / n
    return w, p


def _osc_eigen(N, pnts):
    """
    Vector of and N-dim oscillator eigenfunctions evaluated
    at the points in pnts.
    """
    pnts = np.asarray(pnts)
    lpnts = len(pnts)
    A = np.zeros((N, lpnts))
    A[0, :] = np.exp(-pnts ** 2 / 2.0) / pi ** 0.25
    if N == 1:
        return A
    else:
        A[1, :] = np.sqrt(2) * pnts * A[0, :]
        for k in range(2, N):
            A[k, :] = np.sqrt(2.0 / k) * pnts * A[k - 1, :] - \
                np.sqrt((k - 1.0) / k) * A[k - 2, :]
        return A


def _wigner_clenshaw(rho, xvec, yvec, g=sqrt(2), sparse=False):
    r"""
    Using Clenshaw summation - numerically stable and efficient
    iterative algorithm to evaluate polynomial series.

    The Wigner function is calculated as
    :math:`W = e^(-0.5*x^2)/pi * \sum_{L} c_L (2x)^L / \sqrt(L!)` where
    :math:`c_L = \sum_n \rho_{n,L+n} LL_n^L` where
    :math:`LL_n^L = (-1)^n \sqrt(L!n!/(L+n)!) LaguerreL[n,L,x]`
    """

    M = np.prod(rho.shape[0])
    X,Y = np.meshgrid(xvec, yvec)
    #A = 0.5 * g * (X + 1.0j * Y)
    A2 = g * (X + 1.0j * Y) #this is A2 = 2*A

    B = np.abs(A2)
    B *= B
    w0 = (2*rho.data[0,-1])*np.ones_like(A2)
    L = M-1
    #calculation of \sum_{L} c_L (2x)^L / \sqrt(L!)
    #using Horner's method
    if not sparse:
        rho = rho.full() * (2*np.ones((M,M)) - np.diag(np.ones(M)))
        while L > 0:
            L -= 1
            #here c_L = _wig_laguerre_val(L, B, np.diag(rho, L))
            w0 = _wig_laguerre_val(L, B, np.diag(rho, L)) + w0 * A2 * (L+1)**-0.5
    else:
        while L > 0:
            L -= 1
            diag = _csr_get_diag(rho.data.data,rho.data.indices,
                                rho.data.indptr,L)
            if L != 0:
                diag *= 2
            #here c_L = _wig_laguerre_val(L, B, np.diag(rho, L))
            w0 = _wig_laguerre_val(L, B, diag) + w0 * A2 * (L+1)**-0.5

    return w0.real * np.exp(-B*0.5) * (g*g*0.5 / pi)


def _wig_laguerre_val(L, x, c):
    r"""
    this is evaluation of polynomial series inspired by hermval from numpy.
    Returns polynomial series

    .. math:
        \sum_n b_n LL_n^L,

    where

    .. math:
        LL_n^L = (-1)^n \sqrt(L!n!/(L+n)!) LaguerreL[n,L,x]

    The evaluation uses Clenshaw recursion.
    """

    if len(c) == 1:
        y0 = c[0]
        y1 = 0
    elif len(c) == 2:
        y0 = c[0]
        y1 = c[1]
    else:
        k = len(c)
        y0 = c[-2]
        y1 = c[-1]
        for i in range(3, len(c) + 1):
            k -= 1
            y0,    y1 = c[-i] - y1 * (float((k - 1)*(L + k - 1))/((L+k)*k))**0.5, \
            y0 - y1 * ((L + 2*k -1) - x) * ((L+k)*k)**-0.5

    return y0 - y1 * ((L + 1) - x) * (L + 1)**-0.5


# -----------------------------------------------------------------------------
# Q FUNCTION
#
def _qfunc_check_state(state: Qobj):
    if not isinstance(state, Qobj):
        raise TypeError(f"state must be Qobj, but is {state}")
    # This is only approximate, but it's enough for our purposes; doing more
    # than this would take computational effort we don't _need_ to do.
    isdm = (
        state.isoper
        and state.dims[0] == state.dims[1]
        and state.isherm
        and abs(state.tr() - 1) < qutip.settings.atol
    )
    if not (state.isket or isdm):
        raise ValueError(
            f"state must be a ket or density matrix, but is {state}"
        )
    if len(state.dims[0]) != 1:
        raise ValueError(
            "state must not have tensor structure, but has dimensions:"
            f" {state.dims}"
        )
    return state


def _qfunc_check_coordinates(xvec, yvec):
    if np.isscalar(xvec) or xvec is None:
        raise TypeError("xvec must be array-like, but is " + repr(xvec))
    if np.isscalar(yvec) or yvec is None:
        raise TypeError("yvec must be array-like, but is " + repr(yvec))
    xvec = np.asarray(xvec, dtype=np.float64)
    yvec = np.asarray(yvec, dtype=np.float64)
    if xvec.ndim != 1 or yvec.ndim != 1:
        raise ValueError(
            f"xvec and yvec must be 1D, but have shapes {xvec.shape} and {yvec.shape}."
        )
    return xvec, yvec


class _QFuncCoherentGrid:
    """
    Internal function to compute coherent state operators corresponding to a
    grid of complex values in phase space.  For efficiency reasons, this class
    produces the adjoint of the coherent states, to save allocations when
    calculating inner products later.

    Examples
    --------
    Initialise the grid calculator.

    >>> xvec = yvec = np.linspace(-1, 1, 21)
    >>> g = np.sqrt(0.5)
    >>> max_ns = 10
    >>> grid = _QFuncCoherentGrid(xvec, yvec, g)

    The naive construction of the grid is

    >>> xs, ys = np.meshgrid(xvec, yvec)
    >>> all_alphas = 0.5 * g * (xs + 1j*ys)
    >>> naive = np.array([
    ...     [
    ...         qutip.coherent(max_ns, alpha, method='analytic')
    ...             .dag().full().ravel()
    ...         for alpha in x_alphas
    ...     ]
    ...     for y_alphas in all_alphas
    ... ])

    The naive approach is typically several of orders of magnitude slower than
    this class, which uses much simpler vectorised operations.  The outputs are
    within close tolerance, however:

    >>> np.allclose(naive, grid(max_ns))
    True
    >>> np.allclose(naive[:, :, 4:7], grid(4, 7))
    True
    """
    def __init__(self, xvec, yvec, g: float):
        self.xvec, self.yvec = _qfunc_check_coordinates(xvec, yvec)
        x, y = np.meshgrid(0.5 * g * self.xvec, 0.5 * g * self.yvec)
        self.grid = np.empty(x.shape, dtype=np.complex128)
        self.grid.real = x
        # We produce the adjoint of the coherent states to save an operation
        # later when computing dot products, hence the negative imaginary part.
        self.grid.imag = -y
        self.prefactor = np.exp(-0.5 * (x * x + y * y)).astype(np.complex128)

    def _start(self, first: int):
        """
        Get the coherent state matrix corresponding to the first needed Fock
        state.
        """
        if first == 0:
            return self.prefactor.copy()
        out = np.power(self.grid, first)
        out *= self.prefactor
        return out

    def __call__(self, first: int, last: int = None):
        """
        Get a 3D array of shape ``(yvec.size, xvec.size, last - first)`` of the
        coherent-state vectors for all the Fock states in the range ``first``
        to ``last``, excluding the end point.  The first two axes are the y-
        and x-coordinates of phase space (i.e. Cartesian indexing, like
        ``numpy.meshgrid``), and the last runs over the selected range of
        Fock-space dimensions.
        """
        ns = np.arange(first, last).reshape(1, 1, -1)
        # Technically we could avoid hitting the limits of floating-point
        # exponents for longer by doing all this in logarithmic space (using
        # scipy.special.gammaln), but that ends up involving more
        # floating-point operations overall, and needs special care around the
        # point alpha = 0 to avoid nan appearing, due to how Python handles
        # mixed-width arithmetic operations.
        out = np.empty(self.grid.shape + (ns.size,), dtype=np.complex128)
        out[:, :, 0] = self._start(ns.flat[0])
        for i in range(ns.size - 1):
            out[:, :, i+1] = out[:, :, i] * self.grid
        out /= np.sqrt(scipy.special.factorial(ns))
        return out


class QFunc:
    r"""
    Class-based method of calculating the Husimi-Q function of many different
    quantum states at fixed phase-space points ``0.5*g* (xvec + i*yvec)``.
    This class has slightly higher first-usage costs than :obj:`.qfunc`, but
    subsequent operations will be several times faster. However, it can require
    quite a lot of memory. Call the created object as a function to retrieve
    the Husimi-Q function.

    Parameters
    ----------
    xvec, yvec : array_like
        x- and y-coordinates at which to calculate the Husimi-Q function.

    g : float, default sqrt(2)
        Scaling factor for ``a = 0.5 * g * (x + iy)``.  The value of `g` is
        related to the value of `hbar` in the commutation relation
        :math:`[x,\,y] = i\hbar` via :math:`\hbar=2/g^2`, so the default
        corresponds to :math:`\hbar=1`.

    memory : real, default 1024
        Size in MB that may be used internally as workspace.  This class will
        raise ``MemoryError`` if subsequently passed a state of sufficiently
        large dimension that this bound would be exceeded.  In those cases, use
        :obj:`.qfunc` with ``precompute_memory=None`` instead to force using
        the slower, more memory-efficient algorithm.

    Examples
    --------
    Initialise the class for a square set of coordinates, with some states we
    want to investigate.

    >>> xvec = np.linspace(-2, 2, 101)
    >>> states = [qutip.rand_dm(10) for _ in [None]*10]
    >>> qfunc = qutip.QFunc(xvec, xvec)

    Now we can calculate the Husimi-Q function over each of the states more
    efficiently with:

    >>> husimiq = np.array([qfunc(state) for state in states])

    See Also
    --------
    :obj:`.qfunc` :
        a single function version, which will involve computing several
        quantities multiple times in order to use less memory.
    """

    def __init__(
        self, xvec, yvec, g: float = np.sqrt(2), memory: float = 1024
    ):
        self._g = g
        self._coherent_grid = _QFuncCoherentGrid(xvec, yvec, g)
        # 16 bytes per complex, 1024**2 bytes per MB.
        self._size_mb = self._coherent_grid.grid.size * 16 / (1024 ** 2)
        self._memory_mb = memory
        self._max_size = int(self._memory_mb // self._size_mb)
        self._current_size = 0
        self._cache = None

    def _alphas(self, size: int):
        r"""
        Retrive the full tensor of (the conjugate of) coherent states over all
        values of :math:`\alpha`, for states of dimension ``size``.
        """
        if self._current_size >= size:
            return self._cache[:, :, :size]
        if size > self._max_size:
            requirement = self._size_mb * size
            raise MemoryError(
                f"Refusing to precompute up to {size} basis states."
                f" This would require {requirement:.2f} MB,"
                f" but only {self._memory_mb} MB is allowed."
            )
        if self._cache is None:
            self._cache = self._coherent_grid(self._current_size, size)
        else:
            self._cache = np.dstack(
                [self._cache, self._coherent_grid(self._current_size, size)]
            )
        self._current_size = size
        return self._cache

    def _single(self, vector: np.ndarray, alphas: np.ndarray):
        r"""
        Get the Q function (without the :math:`\pi` scaling factor) of a single
        state vector.
        """
        return np.abs(np.dot(alphas, (self._g * 0.5) * vector)) ** 2

    def __call__(self, state: Qobj):
        """
        Get the Husimi-Q function for the given state vector or density matrix,
        over the coordinates used to initialise the class.  If called multiple
        times, the states do not need to have the same dimensions, but none of
        them can have tensor-product structure.
        """
        state = _qfunc_check_state(state)
        alphas = self._alphas(state.shape[0])
        if state.isket:
            return self._single(state.full().ravel(), alphas) / np.pi
        # We don't use Qobj.eigenstates() to avoid building many unnecessary
        # CSR versions of dense matrices.
        values, vectors = eigh(state.full())
        vectors = vectors.T
        out = values[0] * self._single(vectors[0], alphas)
        for value, vector in zip(values[1:], vectors[1:]):
            out += value * self._single(vector, alphas)
        return out / np.pi


def _qfunc_iterative_single(
    vector: np.ndarray, alpha_grid: _QFuncCoherentGrid, g: float,
):
    r"""
    Get the Q function (without the :math:`\pi` scaling factor) of a single
    state vector, using the iterative algorithm which recomputes the powers of
    the coherent-state matrix.
    """
    ns = np.arange(vector.shape[0])
    out = np.polyval(
        (0.5*g * vector / np.sqrt(scipy.special.factorial(ns)))[::-1],
        alpha_grid.grid,
    )
    out *= alpha_grid.prefactor
    return np.abs(out)**2


def qfunc(
    state: Qobj,
    xvec,
    yvec,
    g: float = sqrt(2),
    precompute_memory: float = 1024,
):
    r"""
    Husimi-Q function of a given state vector or density matrix at phase-space
    points ``0.5 * g * (xvec + i*yvec)``.

    Parameters
    ----------
    state : :obj:`.Qobj`
        A state vector or density matrix.  This cannot have tensor-product
        structure.

    xvec, yvec : array_like
        x- and y-coordinates at which to calculate the Husimi-Q function.

    g : float, default sqrt(2)
        Scaling factor for ``a = 0.5 * g * (x + iy)``.  The value of `g` is
        related to the value of :math:`\hbar` in the commutation relation
        :math:`[x,\,y] = i\hbar` via :math:`\hbar=2/g^2`, so the default
        corresponds to :math:`\hbar=1`.

    precompute_memory : real, default 1024
        Size in MB that may be used during calculations as working space when
        dealing with density-matrix inputs.  This is ignored for state-vector
        inputs.  The bound is not quite exact due to other, order-of-magnitude
        smaller, intermediaries being necessary, but is a good approximation.
        If you want to use the same iterative algorithm for density matrices
        that is used for single kets, set ``precompute_memory=None``.

    Returns
    --------
    ndarray
        Values representing the Husimi-Q function calculated over the specified
        range ``[xvec, yvec]``.

    See Also
    --------
    :obj:`.QFunc` :
        a class-based version, more efficient if you want to calculate the
        Husimi-Q function for several states over the same coordinates.
    """
    state = _qfunc_check_state(state)
    xvec, yvec = _qfunc_check_coordinates(xvec, yvec)
    required_memory = state.shape[0] * xvec.size * yvec.size * 16 / (1024 ** 2)
    enough_memory = (
        precompute_memory is not None
        and precompute_memory > required_memory
    )
    if state.isoper and enough_memory:
        return QFunc(xvec, yvec, g)(state)
    if precompute_memory is not None and state.isoper:
        warnings.warn(
            "Falling back to iterative algorithm due to lack of memory."
            f" Needed {required_memory:.2f} MB, but only allowed to use"
            f" {precompute_memory:.2f} MB.  Increase `precompute_memory` to"
            " raise limit, or set to `None` to suppress warning."
        )
    alpha_grid = _QFuncCoherentGrid(xvec, yvec, g)
    if state.isket:
        out = _qfunc_iterative_single(state.full().ravel(), alpha_grid, g)
        out /= np.pi
        return out
    # We don't use Qobj.eigenstates() to avoid building many unnecessary CSR
    # versions of dense matrices.
    values, vectors = eigh(state.full())
    vectors = vectors.T
    out = values[0] * _qfunc_iterative_single(vectors[0], alpha_grid, g)
    for value, vector in zip(values[1:], vectors[1:]):
        out += value * _qfunc_iterative_single(vector, alpha_grid, g)
    out /= np.pi
    return out


# -----------------------------------------------------------------------------
# PSEUDO DISTRIBUTION FUNCTIONS FOR SPINS
#
def spin_q_function(rho, theta, phi):
    r"""The Husimi Q function for spins is defined as ``Q(theta, phi) =
    SCS.dag() * rho * SCS`` for the spin coherent state ``SCS = spin_coherent(
    j, theta, phi)`` where j is the spin length.
    The implementation here is more efficient as it doesn't
    generate all of the SCS at theta and phi (see references).

    The spin Q function is normal when integrated over the surface of the
    sphere

    .. math:: \frac{4 \pi}{2j + 1}\int_\phi \int_\theta
              Q(\theta, \phi) \sin(\theta) d\theta d\phi = 1

    Parameters
    ----------
    state : qobj
        A state vector or density matrix for a spin-j quantum system.
    theta : array_like
        Polar (colatitude) angle at which to calculate the Husimi-Q function.
    phi : array_like
        Azimuthal angle at which to calculate the Husimi-Q function.

    Returns
    -------
    Q, THETA, PHI : 2d-array
        Values representing the spin Husimi Q function at the values specified
        by THETA and PHI.

    References
    ----------
    [1] Lee Loh, Y., & Kim, M. (2015). American J. of Phys., 83(1), 30–35.
    https://doi.org/10.1119/1.4898595

    """

    if rho.type == 'bra':
        rho = rho.dag()

    if rho.type == 'ket':
        rho = ket2dm(rho)

    J = rho.shape[0]
    j = (J - 1) / 2

    THETA, PHI = meshgrid(theta, phi)

    Q = np.zeros_like(THETA, dtype=complex)

    for m1 in arange(-j, j + 1):
        Q += binom(2 * j, j + m1) * cos(THETA / 2) ** (2 * (j + m1)) * \
             sin(THETA / 2) ** (2 * (j - m1)) * \
             rho.data[int(j - m1), int(j - m1)]

        for m2 in arange(m1 + 1, j + 1):
            Q += (sqrt(binom(2 * j, j + m1)) * sqrt(binom(2 * j, j + m2)) *
                  cos(THETA / 2) ** (2 * j + m1 + m2) *
                  sin(THETA / 2) ** (2 * j - m1 - m2)) * \
             (exp(1j * (m1 - m2) * PHI) * rho.data[int(j - m1), int(j - m2)] +
              exp(1j * (m2 - m1) * PHI) * rho.data[int(j - m2), int(j - m1)])

    return Q.real, THETA, PHI


def _rho_kq(rho, j, k, q):
    """
    This calculates the trace of the multipole operator T_kq and the density
    matrix rho for use in the spin Wigner quasiprobability distribution.

    Parameters
    ----------
    rho : qobj
        A density matrix for a spin-j quantum system.
    j : float
        The spin length of the system.
    k : int
        Spherical harmonic degree
    q : int
        Spherical harmonic order

    Returns
    -------
    v : float
        Overlap of state with multipole operator T_kq
    """

    v = 0j
    for m1 in arange(-j, j+1):
        for m2 in arange(-j, j+1):
            v += (
                    (-1) ** (2 * j - k - m1 - m2)
                    * np.sqrt((2 * k + 1) / (2 * j + 1))
                    * qutip.clebsch(j, k, j, -m1, q, -m2)
                    * rho.data[int(j - m1), int(j - m2)]
            )
    return v


def spin_wigner(rho, theta, phi):
    r"""Wigner function for a spin-j system.

    The spin W function is normal when integrated over the surface of the
    sphere

    .. math:: \sqrt{\frac{4 \pi}{2j + 1}}\int_\phi \int_\theta
              W(\theta,\phi) \sin(\theta) d\theta d\phi = 1


    Parameters
    ----------
    state : qobj
        A state vector or density matrix for a spin-j quantum system.
    theta : array_like
        Polar (colatitude) angle at which to calculate the W function.
    phi : array_like
        Azimuthal angle at which to calculate the W function.

    Returns
    -------
    W, THETA, PHI : 2d-array
        Values representing the spin Wigner function at the values specified
        by THETA and PHI.

    References
    ----------
    [1] Agarwal, G. S. (1981). Phys. Rev. A, 24(6), 2889–2896.
    https://doi.org/10.1103/PhysRevA.24.2889

    [2] Dowling, J. P., Agarwal, G. S., & Schleich, W. P. (1994).
    Phys. Rev. A, 49(5), 4101–4109. https://doi.org/10.1103/PhysRevA.49.4101

    [3] Conversion between Wigner 3-j symbol and Clebsch-Gordan coefficients
    taken from Wikipedia (https://en.wikipedia.org/wiki/3-j_symbol)

    """

    if rho.type == 'bra':
        rho = rho.dag()

    if rho.type == 'ket':
        rho = ket2dm(rho)

    J = rho.shape[0]
    j = (J - 1) / 2

    THETA, PHI = meshgrid(theta, phi)

    W = np.zeros_like(THETA, dtype=complex)

    for k in range(int(2 * j)+1):
        for q in arange(-k, k+1):
            # sph_harm takes azimuthal angle then polar angle as arguments
            W += _rho_kq(rho, j, k, q) * sph_harm(q, k, PHI, THETA)

    return W.real, THETA, PHI



# This file is part of QuTiP: Quantum Toolbox in Python.
#
#    Copyright (c) 2011 and later, Paul D. Nation and Robert J. Johansson.
#    All rights reserved.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions are
#    met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of the QuTiP: Quantum Toolbox in Python nor the names
#       of its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
###############################################################################

import pytest
import collections
import functools
import numpy as np
import qutip

# We want to test the broadcasting rules for `qutip.expect` for a whole bunch
# of different systems, without having to repeatedly specify the systems over
# and over again.  We first store a small number of test cases for known
# expectation value in the most bundled-up form, because it's easier to unroll
# these by applying the expected broadcasting rules explicitly ourselves than
# performing the inverse operation.
#
# We store a single test case in a record type, just to keep things neatly
# together while we're munging them, so it's clear at all times what
# constitutes a valid test case.

_Case = collections.namedtuple('_Case', ['operator', 'state', 'expected'])


def _case_to_dm(case):
    return case._replace(state=[x.proj() for x in case.state])


def _unwrap(list_):
    """Unwrap lists until we reach the first non-list element."""
    out = list_
    while isinstance(out, list):
        out = out[0]
    return out


def _case_id(case):
    op_part = 'qubit' if _unwrap(case.operator).dims[0][0] == 2 else 'basis'
    state_part = 'ket' if _unwrap(case.state).dims[1][0] == 1 else 'dm'
    return op_part + "-" + state_part


# This is the minimal set of test cases, with a Fock system and a qubit system
# both in ket form and dm form.  The reference expectations are a 2D array
# which would be found by broadcasting `operator` against `state` and applying
# `qutip.expect` to the pairs.
_dim = 5
_num, _a = qutip.num(_dim), qutip.destroy(_dim)
_sx, _sz, _sp = qutip.sigmax(), qutip.sigmaz(), qutip.sigmap()
_known_fock = _Case([_num, _a],
                    [qutip.fock(_dim, n) for n in range(_dim)],
                    np.array([np.arange(_dim), np.zeros(_dim)]))
_known_qubit = _Case([_sx, _sz, _sp],
                     [qutip.basis(2, 0), qutip.basis(2, 1)],
                     np.array([[0, 0], [1, -1], [0, 0]]))
_known_cases = [_known_fock, _case_to_dm(_known_fock),
                _known_qubit, _case_to_dm(_known_qubit)]


class TestKnownExpectation:
    def pytest_generate_tests(self, metafunc):
        """
        Perform the parametrisation over the test cases, performing the
        explicit broadcasting into separate test cases when required.

        We detect whether to perform explicit broadcasting over one of the
        arguments of the `_Case` by looking for a singular/plural name of the
        parameter in the test.  If the parameter is singular, then we manually
        perform the broadcasting rule for that fixture, and parametrise over
        the resulting list, taking care to pick out the correct parts of the
        reference array.
        """
        cases = _known_cases
        op_name, state_name = 'operator', 'state'
        if op_name not in metafunc.fixturenames:
            op_name += 's'
        else:
            cases = [_Case(op, case.state, expected)
                     for case in cases
                     for op, expected in zip(case.operator, case.expected)]
        if state_name not in metafunc.fixturenames:
            state_name += 's'
        else:
            cases = [_Case(case.operator, state, expected)
                     for case in cases
                     for state, expected in zip(case.state, case.expected.T)]
        metafunc.parametrize([op_name, state_name, 'expected'], cases,
                             ids=[_case_id(case) for case in cases])

    def test_operator_by_basis(self, operator, state, expected):
        result = qutip.expect(operator, state)
        assert result == expected
        assert isinstance(result, float if operator.isherm else complex)

    def test_broadcast_operator_list(self, operators, state, expected):
        result = qutip.expect(operators, state)
        expected_dtype = (np.float64 if all(op.isherm for op in operators)
                          else np.complex128)
        assert isinstance(result, np.ndarray)
        assert result.dtype == expected_dtype
        assert list(result) == list(expected)

    def test_broadcast_state_list(self, operator, states, expected):
        result = qutip.expect(operator, states)
        expected_dtype = np.float64 if operator.isherm else np.complex128
        assert isinstance(result, np.ndarray)
        assert result.dtype == expected_dtype
        assert list(result) == list(expected)

    def test_broadcast_both_lists(self, operators, states, expected):
        result = qutip.expect(operators, states)
        assert len(result) == len(operators)
        for part, operator, expected_part in zip(result, operators, expected):
            expected_dtype = np.float64 if operator.isherm else np.complex128
            assert isinstance(part, np.ndarray)
            assert part.dtype == expected_dtype
            assert list(part) == list(expected_part)


@pytest.mark.repeat(20)
@pytest.mark.parametrize("hermitian", [False, True], ids=['complex', 'real'])
def test_equivalent_to_matrix_element(hermitian):
    dimension = 20
    state = qutip.rand_ket(dimension, 0.3)
    op = qutip.rand_herm(dimension, 0.2)
    if not hermitian:
        op = op + 1j*qutip.rand_herm(dimension, 0.1)
    expected = (state.dag() * op * state).data[0, 0]
    assert abs(qutip.expect(op, state) - expected) < 1e-14


@pytest.mark.parametrize("solve", [
    pytest.param(qutip.sesolve, id="sesolve"),
    pytest.param(functools.partial(qutip.mesolve, c_ops=[qutip.qzero(2)]),
                 id="mesolve"),
])
def test_compatibility_with_solver(solve):
    e_ops = [getattr(qutip, 'sigma'+x)() for x in 'xyzmp']
    h = qutip.sigmax()
    state = qutip.basis(2, 0)
    times = np.linspace(0, 10, 101)
    options = qutip.Options(store_states=True)
    result = solve(h, state, times, e_ops=e_ops, options=options)
    direct, states = result.expect, result.states
    indirect = qutip.expect(e_ops, states)
    assert len(direct) == len(indirect)
    for direct_, indirect_ in zip(direct, indirect):
        assert len(direct_) == len(indirect_)
        assert isinstance(direct_, np.ndarray)
        assert isinstance(indirect_, np.ndarray)
        assert direct_.dtype == indirect_.dtype
        np.testing.assert_allclose(direct_, indirect_, atol=1e-12)



# -*- coding: utf-8 -*-
# This file is part of QuTiP: Quantum Toolbox in Python.
#
#    Copyright (c) 2014 and later, Alexander J G Pitchford
#    All rights reserved.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions are
#    met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of the QuTiP: Quantum Toolbox in Python nor the names
#       of its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
###############################################################################

# @author: Alexander Pitchford
# @email1: agp1@aber.ac.uk
# @email2: alex.pitchford@gmail.com
# @organization: Aberystwyth University
# @supervisor: Daniel Burgarth
# @date: Sep 2015

import pytest
import collections
import os
import pathlib
import tempfile
import numpy as np
import scipy.optimize
import scipy.sparse as sp

import qutip
from qutip.control import pulseoptim as cpo
from qutip.qip.algorithms import qft
from qutip.qip.operations.gates import hadamard_transform
import qutip.control.loadparams

_sx = qutip.sigmax()
_sy = qutip.sigmay()
_sz = qutip.sigmaz()
_sp = qutip.sigmap()
_sm = qutip.sigmam()
_si = qutip.identity(2)
_project_0 = qutip.basis(2, 0).proj()
_hadamard = hadamard_transform(1)

# We have a whole bunch of different physical systems we want to test the
# optimiser for, but the logic for testing them is largely the same.  To avoid
# having to explicitly parametrise over five linked parameters repeatedly, we
# group them into a record type, so that all the optimisation functions can
# then simply be parametrised over a single argument.
#
# We supply `kwargs` as a property of the system because the initial pulse type
# and dynamics solver to use vary, especially if the system is unitary.
_System = collections.namedtuple('_System',
                                 ['system', 'controls', 'initial', 'target',
                                  'kwargs'])

# Simple Hadamard gate.
_hadamard_kwargs = {'num_tslots': 10, 'evo_time': 10, 'gen_stats': True,
                    'init_pulse_type': 'LIN', 'fid_err_targ': 1e-10,
                    'dyn_type': 'UNIT'}
hadamard = _System(system=_sz,
                   controls=[_sx],
                   initial=_si,
                   target=_hadamard,
                   kwargs=_hadamard_kwargs)

# Quantum Fourier transform.
_qft_system = 0.5 * sum(qutip.tensor(op, op) for op in (_sx, _sy, _sz))
_qft_controls = [0.5*qutip.tensor(_sx, _si), 0.5*qutip.tensor(_sy, _si),
                 0.5*qutip.tensor(_si, _sx), 0.5*qutip.tensor(_si, _sy)]
_qft_kwargs = {'num_tslots': 10, 'evo_time': 10, 'gen_stats': True,
               'init_pulse_type': 'LIN', 'fid_err_targ': 1e-9,
               'dyn_type': 'UNIT'}
qft = _System(system=_qft_system,
              controls=_qft_controls,
              initial=qutip.identity([2, 2]),
              target=qft.qft(2),
              kwargs=_qft_kwargs)

# Coupling constants are completely arbitrary.
_ising_system = (0.9*qutip.tensor(_sx, _si) + 0.7*qutip.tensor(_si, _sx)
                 + 0.8*qutip.tensor(_sz, _si) + 0.9*qutip.tensor(_si, _sz))
_ising_kwargs = {'num_tslots': 10, 'evo_time': 18, 'init_pulse_type': 'LIN',
                 'fid_err_targ': 1e-10, 'dyn_type': 'UNIT'}
ising = _System(system=_ising_system,
                controls=[qutip.tensor(_sz, _sz)],
                initial=qutip.basis([2, 2], [0, 0]),
                target=qutip.basis([2, 2], [1, 1]),
                kwargs=_ising_kwargs)

# Louivillian amplitude-damping channel system.
_l_adc_system = 0.1 * (2*qutip.tensor(_sm, _sp.dag())
                       - qutip.tensor(_project_0, _si)
                       - qutip.tensor(_si, _project_0.dag()))
_l_adc_controls = [1j * (qutip.tensor(_si, _sz) - qutip.tensor(_sz, _si)),
                   1j * (qutip.tensor(_si, _sx) - qutip.tensor(_sx, _si))]
_l_adc_kwargs = {'num_tslots': 10, 'evo_time': 5, 'init_pulse_type': 'LIN',
                 'max_iter': 200, 'fid_err_targ': 1e-1, 'gen_stats': True}
l_adc = _System(system=_l_adc_system,
                controls=_l_adc_controls,
                initial=qutip.identity([2, 2]),
                target=hadamard_transform(2),
                kwargs=_l_adc_kwargs)

# Two coupled oscillators with symplectic dynamics.
_g1, _g2 = 1.0, 0.2
_A_rotate = qutip.qdiags([[1, 1, 0, 0]], [0])
_A_squeeze = 0.4 * qutip.qdiags([[1, -1, 0, 0]], [0])
_A_target = qutip.qdiags([[1, 1], [1, 1]], [2, -2])
_Omega = qutip.Qobj(qutip.control.symplectic.calc_omega(2))
_sympl_system = qutip.qdiags([[1, 1, 1, 1], [_g1, _g2], [_g1, _g2]],
                             [0, 2, -2])
_sympl_target = (-0.5 * _A_target * _Omega * np.pi).expm()
_sympl_kwargs = {'num_tslots': 20, 'evo_time': 10, 'fid_err_targ': 1e-3,
                 'max_iter': 200, 'dyn_type': 'SYMPL',
                 'init_pulse_type': 'ZERO', 'gen_stats': True}
symplectic = _System(system=_sympl_system,
                     controls=[_A_rotate, _A_squeeze],
                     initial=qutip.identity(4),
                     target=_sympl_target,
                     kwargs=_sympl_kwargs)


# Parametrise the systems and the propagation method separately so that we test
# all combinations of both.

# Test propagation with the default settings and with internal Qobj use for all
# test cases.
@pytest.fixture(params=[
    pytest.param(None, id="default propagation"),
    pytest.param({'oper_dtype': qutip.Qobj}, id="Qobj propagation"),
    pytest.param({'oper_dtype': np.ndarray}, id="ndarray propagation"),
])
def propagation(request):
    return {'dyn_params': request.param}


# Any test requiring a system to test will parametrise over all of the ones we
# defined above.
@pytest.fixture(params=[
    pytest.param(hadamard, id="Hadamard gate"),
    pytest.param(qft, id="QFT"),
    pytest.param(ising, id="Ising state-to-state"),
    pytest.param(l_adc, id="Lindbladian amplitude damping channel"),
    pytest.param(symplectic, id="Symplectic coupled oscillators"),
])
def system(request):
    return request.param


def _optimize_pulse(system):
    """
    Unpack the `system` record type, optimise the result and assert that it
    succeeded.
    """
    result = cpo.optimize_pulse(system.system, system.controls,
                                system.initial, system.target,
                                **system.kwargs)
    error = " ".join(["Infidelity: {:7.4e}".format(result.fid_err),
                      "reason:", result.termination_reason])
    assert result.goal_achieved, error
    return result


def _merge_kwargs(system, kwargs):
    """
    Return a copy of `system` with any passed `kwargs` updated in the
    dictionary---this can be used to overwrite or to add new arguments.
    """
    out = system.kwargs.copy()
    out.update(kwargs)
    return system._replace(kwargs=out)


class TestOptimization:
    def test_basic_optimization(self, system, propagation):
        """Test the optimiser in the base case for each system."""
        system = _merge_kwargs(system, propagation)
        result = _optimize_pulse(system)
        assert result.fid_err < system.kwargs['fid_err_targ']

    def test_sparse_eigen_optimization(self, system, propagation):
        system = _merge_kwargs(system, {
            **propagation,
            "dyn_params": {
                **(propagation.get("dyn_params") or {}),
                "sparse_eigen_decomp": True,
            }
        })
        if system.system.dims == [[2], [2]]:
            with pytest.raises(ValueError, match=(
                    r"^Single qubit pulse optimization dynamics cannot use"
                    r" sparse eigenvector decomposition")):
                _optimize_pulse(system)
        else:
            result = _optimize_pulse(system)
            assert result.fid_err < system.kwargs['fid_err_targ']

    @pytest.mark.parametrize("oper_dtype", [
        pytest.param(sp.csr_matrix, id="csr_matrix"),
        pytest.param(list, id="list"),
    ])
    def test_invalid_oper_dtype(self, system, oper_dtype):
        system = _merge_kwargs(system, {"dyn_params": {
            "oper_dtype": oper_dtype,
        }})
        with pytest.raises(ValueError) as err:
            _optimize_pulse(system)
        assert str(err.value) == (
            f"Unknown oper_dtype {oper_dtype!r}. The oper_dtype may be"
            " qutip.Qobj or numpy.ndarray."
        )

    def test_object_oriented_approach_and_gradient(self, system, propagation):
        """
        Test the object-oriented version of the optimiser, and ensure that the
        system truly appears to be at an extremum.
        """
        system = _merge_kwargs(system, propagation)
        base = _optimize_pulse(system)
        optimizer = cpo.create_pulse_optimizer(system.system, system.controls,
                                               system.initial, system.target,
                                               **system.kwargs)
        init_amps = np.array([optimizer.pulse_generator.gen_pulse()
                              for _ in system.controls]).T
        optimizer.dynamics.initialize_controls(init_amps)
        # Check the gradient numerically.
        func = optimizer.fid_err_func_wrapper
        grad = optimizer.fid_err_grad_wrapper
        loc = optimizer.dynamics.ctrl_amps.flatten()
        assert abs(scipy.optimize.check_grad(func, grad, loc)) < 1e-5,\
            "Gradient outside tolerance."
        result = optimizer.run_optimization()
        tol = system.kwargs['fid_err_targ']
        assert abs(result.fid_err-base.fid_err) < tol,\
            "Direct and indirect methods produce different results."

    @pytest.mark.parametrize("kwargs", [
        pytest.param({'gen_stats': False}, id="no stats"),
        pytest.param({'num_tslots': None, 'evo_time': None,
                      'tau': np.arange(1, 10, 1, dtype=np.float64)},
                     id="tau array")
    ])
    def test_modified_optimization(self, propagation, kwargs):
        """Test a basic system with a few different combinations of options."""
        system = _merge_kwargs(hadamard, kwargs)
        self.test_basic_optimization(system, propagation)

    def test_optimizer_bounds(self):
        """Test that bounds on the control fields are obeyed."""
        bound = 1.0
        kwargs = {'amp_lbound': -bound, 'amp_ubound': bound}
        system = _merge_kwargs(qft, kwargs)
        result = _optimize_pulse(system)
        assert np.all(result.final_amps >= -bound)
        assert np.all(result.final_amps <= bound)

    def test_unitarity_via_dump(self):
        """
        Test that unitarity is maintained at all times throughout the
        optimisation of the controls.
        """
        kwargs = {'num_tslots': 1000, 'evo_time': 4, 'fid_err_targ': 1e-9,
                  'dyn_params': {'dumping': 'FULL'}}
        system = _merge_kwargs(hadamard, kwargs)
        result = _optimize_pulse(system)
        dynamics = result.optimizer.dynamics
        assert dynamics.dump is not None, "Dynamics dump not created"
        # Use the dump to check unitarity of all propagators and evo_ops
        dynamics.unitarity_tol = 1e-13 # 1e-14 for eigh but 1e-13 for eig.
        for item, description in [('prop', 'propagators'),
                                  ('fwd_evo', 'forward evolution operators'),
                                  ('onto_evo', 'onto evolution operators')]:
            non_unitary = sum(not dynamics._is_unitary(x)
                              for dump in dynamics.dump.evo_dumps
                              for x in getattr(dump, item))
            assert non_unitary == 0, "Found non-unitary " + description + "."

    def test_crab(self, propagation):
        tol = 1e-5
        evo_time = 10
        result = cpo.opt_pulse_crab_unitary(
            hadamard.system, hadamard.controls,
            hadamard.initial, hadamard.target,
            num_tslots=12, evo_time=evo_time, fid_err_targ=tol,
            **propagation,
            alg_params={'crab_pulse_params': {'randomize_coeffs': False,
                                              'randomize_freqs': False}},
            init_coeff_scaling=0.5,
            guess_pulse_type='GAUSSIAN',
            guess_pulse_params={'variance': evo_time * 0.1},
            guess_pulse_scaling=1.0,
            guess_pulse_offset=1.0,
            amp_lbound=None,
            amp_ubound=None,
            ramping_pulse_type='GAUSSIAN_EDGE',
            ramping_pulse_params={'decay_time': evo_time * 0.01},
            gen_stats=True)
        error = " ".join(["Infidelity: {:7.4e}".format(result.fid_err),
                          "reason:", result.termination_reason])
        assert result.goal_achieved, error
        assert abs(result.fid_err) < tol
        assert abs(result.final_amps[0, 0]) < tol, "Lead-in amplitude nonzero."


# The full object-orientated interface to the optimiser is rather complex.  To
# attempt to simplify the test of the configuration loading, we break it down
# into steps here.

def _load_configuration(path):
    configuration = qutip.control.optimconfig.OptimConfig()
    configuration.param_fname = path.name
    configuration.param_fpath = str(path)
    configuration.pulse_type = "ZERO"
    qutip.control.loadparams.load_parameters(str(path), config=configuration)
    return configuration


def _load_dynamics(path, system, configuration, stats):
    dynamics = qutip.control.dynamics.DynamicsUnitary(configuration)
    dynamics.drift_dyn_gen = system.system
    dynamics.ctrl_dyn_gen = system.controls
    dynamics.initial = system.initial
    dynamics.target = system.target
    qutip.control.loadparams.load_parameters(str(path), dynamics=dynamics)
    dynamics.init_timeslots()
    dynamics.stats = stats
    return dynamics


def _load_pulse_generator(path, configuration, dynamics):
    pulse_generator = qutip.control.pulsegen.create_pulse_gen(
            pulse_type=configuration.pulse_type,
            dyn=dynamics)
    qutip.control.loadparams.load_parameters(str(path),
                                             pulsegen=pulse_generator)
    return pulse_generator


def _load_termination_conditions(path):
    conditions = qutip.control.termcond.TerminationConditions()
    qutip.control.loadparams.load_parameters(str(path), term_conds=conditions)
    return conditions


def _load_optimizer(path, configuration, dynamics, pulse_generator,
                    termination_conditions, stats):
    method = configuration.optim_method
    if method is None:
        raise qutip.control.errors.UsageError(
            "Optimization algorithm must be specified using the 'optim_method'"
            " parameter.")
    known = {'BFGS': 'OptimizerBFGS', 'FMIN_L_BFGS_B': 'OptimizerLBFGSB'}
    constructor = getattr(qutip.control.optimizer,
                          known.get(method, 'Optimizer'))
    optimizer = constructor(configuration, dynamics)
    optimizer.method = method
    qutip.control.loadparams.load_parameters(str(path), optim=optimizer)
    optimizer.config = configuration
    optimizer.dynamics = dynamics
    optimizer.pulse_generator = pulse_generator
    optimizer.termination_conditions = termination_conditions
    optimizer.stats = stats
    return optimizer


class TestFileIO:
    def test_load_parameters_from_file(self):
        system = hadamard
        path = pathlib.Path(__file__).parent / "Hadamard_params.ini"
        stats = qutip.control.stats.Stats()
        configuration = _load_configuration(path)
        dynamics = _load_dynamics(path, system, configuration, stats)
        pulse_generator = _load_pulse_generator(path, configuration, dynamics)
        termination_conditions = _load_termination_conditions(path)
        optimizer = _load_optimizer(path,
                                    configuration,
                                    dynamics,
                                    pulse_generator,
                                    termination_conditions,
                                    stats)
        init_amps = np.array([optimizer.pulse_generator.gen_pulse()
                              for _ in system.controls]).T
        optimizer.dynamics.initialize_controls(init_amps)
        result = optimizer.run_optimization()

        kwargs = {'num_tslots': 6, 'evo_time': 6, 'fid_err_targ': 1e-10,
                  'init_pulse_type': 'LIN', 'dyn_type': 'UNIT',
                  'amp_lbound': -1, 'amp_ubound': 1,
                  'gen_stats': True}
        target = _optimize_pulse(system._replace(kwargs=kwargs))
        np.testing.assert_allclose(result.final_amps, target.final_amps,
                                   atol=1e-5)

    @pytest.mark.usefixtures("in_temporary_directory")
    def test_dumping_to_files(self):
        N_OPTIMDUMP_FILES = 10
        N_DYNDUMP_FILES = 49
        dumping = {'dumping': 'FULL', 'dump_to_file': True}
        kwargs = {'num_tslots': 1_000, 'evo_time': 4, 'fid_err_targ': 1e-9,
                  'optim_params': {'dump_dir': 'optim', **dumping},
                  'dyn_params': {'dump_dir': 'dyn', **dumping}}
        system = _merge_kwargs(hadamard, kwargs)
        result = _optimize_pulse(system)

        # Check dumps were generated and have the right number of files.
        assert result.optimizer.dump is not None
        assert result.optimizer.dynamics.dump is not None
        assert (len(os.listdir(result.optimizer.dump.dump_dir))
                == N_OPTIMDUMP_FILES)
        assert (len(os.listdir(result.optimizer.dynamics.dump.dump_dir))
                == N_DYNDUMP_FILES)

        # Dump all to specific file stream.
        for dump, type_ in [(result.optimizer.dump, 'optimizer'),
                            (result.optimizer.dynamics.dump, 'dynamics')]:
            with tempfile.NamedTemporaryFile() as file:
                dump.writeout(file)
                assert os.stat(file.name).st_size > 0,\
                    " ".join(["Empty", type_, "file."])


def _count_waves(system):
    optimizer = cpo.create_pulse_optimizer(system.system, system.controls,
                                           system.initial, system.target,
                                           **system.kwargs)
    pulse = optimizer.pulse_generator.gen_pulse()
    zero_crossings = pulse[0:-2]*pulse[1:-1] < 0
    return (sum(zero_crossings) + 1) // 2


@pytest.mark.parametrize('pulse_type',
                         [pytest.param(x, id=x.lower())
                          for x in ['SINE', 'SQUARE', 'TRIANGLE', 'SAW']])
class TestPeriodicControlFunction:
    num_tslots = 1_000
    evo_time = 10

    @pytest.mark.parametrize('n_waves', [1, 5, 10, 100])
    def test_number_of_waves(self, pulse_type, n_waves):
        kwargs = {'num_tslots': self.num_tslots, 'evo_time': self.evo_time,
                  'init_pulse_type': pulse_type,
                  'init_pulse_params': {'num_waves': n_waves},
                  'gen_stats': False}
        system = _merge_kwargs(hadamard, kwargs)
        assert _count_waves(system) == n_waves

    @pytest.mark.parametrize('frequency', [0.1, 1, 10, 20])
    def test_frequency(self, pulse_type, frequency):
        kwargs = {'num_tslots': self.num_tslots, 'evo_time': self.evo_time,
                  'init_pulse_type': pulse_type,
                  'init_pulse_params': {'freq': frequency},
                  'fid_err_targ': 1e-5,
                  'gen_stats': False}
        system = _merge_kwargs(hadamard, kwargs)
        assert _count_waves(system) == self.evo_time*frequency


class TestTimeDependence:
    """
    Test that systems where the system Hamiltonian is time-dependent behave as
    expected under the optimiser.
    """
    def test_drift(self):
        """
        Test that introducing time dependence to the system does change the
        result of the optimisation.
        """
        num_tslots = 20
        system = _merge_kwargs(hadamard, {'num_tslots': num_tslots,
                                          'evo_time': 10})
        result_fixed = _optimize_pulse(system)
        system_flat = system._replace(system=[system.system]*num_tslots)
        result_flat = _optimize_pulse(system_flat)
        step = [0.0]*(num_tslots//2) + [1.0]*(num_tslots//2)
        system_step = system._replace(system=[x*system.system for x in step])
        result_step = _optimize_pulse(system_step)
        np.testing.assert_allclose(result_fixed.final_amps,
                                   result_flat.final_amps,
                                   rtol=1e-9)
        assert np.any((result_flat.final_amps-result_step.final_amps) > 1e-3),\
            "Flat and step drights result in the same control pulses."

    def test_controls_all_time_slots_equal_to_no_time_dependence(self):
        """
        Test that simply duplicating the system in each time slot (i.e. no
        actual time dependence has no effect on the final result.
        """
        num_tslots = 20
        system = _merge_kwargs(hadamard, {'num_tslots': num_tslots,
                                          'evo_time': 10,
                                          'fid_err_targ': 1e-10})
        result_single = _optimize_pulse(system)
        system_vary = system._replace(controls=[[_sx]]*num_tslots)
        result_vary = _optimize_pulse(system_vary)
        np.testing.assert_allclose(result_single.final_amps,
                                   result_vary.final_amps,
                                   atol=1e-9)

    def test_controls_identity_operators_ignored(self):
        """
        Test that moments in time where the control parameters are simply the
        identity are just ignored by the optimiser (since they'll never be able
        to do anything.
        """
        num_tslots = 20
        controls = [[_sx] if k % 3 else [_si] for k in range(num_tslots)]
        system = _merge_kwargs(hadamard, {'num_tslots': num_tslots,
                                          'evo_time': 10})
        system = system._replace(controls=controls)
        result = _optimize_pulse(system)
        for k in range(0, num_tslots, 3):
            np.testing.assert_allclose(result.initial_amps[k],
                                       result.final_amps[k],
                                       rtol=1e-9)



# This file is part of QuTiP: Quantum Toolbox in Python.
#
#    Copyright (c) 2011 and later, Paul D. Nation and Robert J. Johansson.
#    All rights reserved.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions are
#    met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of the QuTiP: Quantum Toolbox in Python nor the names
#       of its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
###############################################################################
import functools
import numpy as np
import scipy.interpolate
import pytest
import qutip

pytestmark = [pytest.mark.usefixtures("in_temporary_directory")]


def pytest_generate_tests(metafunc):
    """
    Perform more complex parametrisation logic for tests in this module.  This
    was originally needed because the `brmesolve` solver cannot mix splines and
    functions in the Hamiltonian (it can only accept string-formatted
    time-dependence), but we want to parametrize this for all the other tests.
    """
    if ("solver" in metafunc.fixturenames
            and "coefficients" in metafunc.fixturenames):
        _parametrize_solver_coefficients(metafunc)


@pytest.mark.parametrize('noise', [
    pytest.param(lambda n: 0.1*np.random.rand(n), id="real"),
    pytest.param(lambda n: 0.1j*np.random.rand(n), id="complex"),
])
def test_equivalence_to_scipy(noise):
    x = np.linspace(0, 2*np.pi, 200)
    y = np.sin(x) + noise(x.shape[0])
    test = qutip.Cubic_Spline(x[0], x[-1], y)
    expected = scipy.interpolate.CubicSpline(x, y, bc_type="natural")
    # We use the bc_type="natural", i.e. zero second-derivatives at the
    # boundaries because that matches qutip.
    np.testing.assert_allclose(test(x), expected(x), atol=1e-10)
    centres = 0.5 * (x[:-1] + x[1:])
    # Test some points not in the original.
    for point in centres:
        assert np.abs(test(point) - expected(point)) < 1e-10, "Scalar argument"


class _Case:
    """
    Represents a time-dependent coefficient test case for
    test_usage_in_solvers.  The functions `function`, `spline` and `string`
    return a tuple of (reference, test), where the reference is a function type
    and the test is the type of whatever was requested.  This lets us build up
    test cases where we can get the same coefficient in different formats, so
    the tests can be parametrised.
    """
    def __init__(self, amplitude, function, string):
        self.amplitude = amplitude
        self._function = function
        self.string_coefficient = string

    def __call__(self, t, *_):
        return self.amplitude * self._function(t)

    def function(self):
        return self, self

    def spline(self, times):
        return self, qutip.Cubic_Spline(times[0], times[-1], self(times))

    def string(self):
        return self, self.string_coefficient


_real_cos = _Case(0.25, np.cos, '0.25*cos(t)')
_real_sin = _Case(0.25, np.sin, '0.25*sin(t)')
_complex_pos = _Case(0.1j, np.sin, '0.1j*sin(t)')
_complex_neg = _Case(-0.1j, np.sin, '-0.1j*sin(t)')


def _parametrize_solver_coefficients(metafunc):
    """
    Perform the parametrisation for test cases using a solver and some
    time-dependent coefficients.  This is necessary because not all solvers can
    accept all combinations of time-dependence specifications.
    """
    size = 10
    times = np.linspace(0, 5, 50)
    c_ops = [qutip.qeye(size)]
    solvers = [
        (qutip.sesolve, 'sesolve'),
        (functools.partial(qutip.mesolve, c_ops=c_ops), 'mesolve'),
        (functools.partial(qutip.mcsolve, c_ops=c_ops), "mcsolve"),
        (qutip.brmesolve, 'brmesolve'),
    ]
    # A list of (reference, test) pairs, where the reference is in function
    # form because that's the fastest, and the test is named type.
    coefficients = [
        ([_real_cos.spline(times)], "real-spline"),
        ([_real_cos.string(), _real_sin.spline(times)], "real-string,spline"),
        ([_real_cos.spline(times), _real_sin.function()],
            "real-spline,function"),
        ([_complex_pos.spline(times), _complex_neg.string()],
            "complex-spline,string"),
    ]

    def _valid_case(solver_id, coefficient_id):
        invalids = [
            "brmesolve" in solver_id and "function" in coefficient_id,
        ]
        return not any(invalids)

    def _make_case(solver, solver_id, coefficients, coefficient_id):
        marks = []
        if "brmesolve" in solver_id:
            # We must use the string as the reference type instead of the
            # function if brmesolve is in use.
            marks.append(pytest.mark.requires_cython)
            coefficients = [(c.string_coefficient, other)
                            for c, other in coefficients]
        id_ = "-".join([solver_id, coefficient_id])
        return pytest.param(solver, coefficients, size, times,
                            id=id_, marks=marks)

    cases = [_make_case(solver, solver_id, coefficients_, coefficient_id)
             for solver, solver_id in solvers
             for coefficients_, coefficient_id in coefficients
             if _valid_case(solver_id, coefficient_id)]
    metafunc.parametrize(["solver", "coefficients", "size", "times"], cases)


@pytest.mark.slow
def test_usage_in_solvers(solver, coefficients, size, times):
    """
    Test that the Cubic_Spline can be used as a time-dependent argument to all
    three principle solvers, both alone and in combination with other forms of
    time-dependence.
    """
    a = qutip.destroy(size)
    state = qutip.basis(size, 1)
    e_ops = [a.dag() * a]
    H_expected = [a.dag()*a]
    H_test = H_expected.copy()
    operator = a**2 + a.dag()**2
    for expected, test in coefficients:
        H_expected.append([operator, expected])
        H_test.append([operator, test])
    expected = solver(H_expected, state, times, e_ops=e_ops).expect[0]
    test = solver(H_test, state, times, e_ops=e_ops).expect[0]
    np.testing.assert_allclose(test, expected, atol=1e-4)



# This file is part of QuTiP: Quantum Toolbox in Python.
#
#    Copyright (c) 2011 and later, Paul D. Nation and Robert J. Johansson.
#    All rights reserved.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions are
#    met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of the QuTiP: Quantum Toolbox in Python nor the names
#       of its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY np.exp(RESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
###############################################################################

import numpy as np
from numpy.testing import run_module_suite, assert_equal
import pytest

from qutip import (
    sigmax, sigmay, sigmaz, sigmam, mesolve, mcsolve, essolve, basis,
)


def _qubit_integrate(tlist, psi0, epsilon, delta, g1, g2, solver):

    H = epsilon / 2.0 * sigmaz() + delta / 2.0 * sigmax()

    c_op_list = []

    rate = g1
    if rate > 0.0:
        c_op_list.append(np.sqrt(rate) * sigmam())

    rate = g2
    if rate > 0.0:
        c_op_list.append(np.sqrt(rate) * sigmaz())

    e_ops = [sigmax(), sigmay(), sigmaz()]

    if solver == "me":
        output = mesolve(H, psi0, tlist, c_op_list, e_ops)
    elif solver == "es":
        output = essolve(H, psi0, tlist, c_op_list, e_ops)
    elif solver == "mc":
        output = mcsolve(H, psi0, tlist, c_op_list, e_ops, ntraj=750)
    else:
        raise ValueError("unknown solver")

    return output.expect[0], output.expect[1], output.expect[2]


def test_MESolverCase1():
    """
    Test mesolve qubit, with dissipation
    """

    epsilon = 0.0 * 2 * np.pi   # cavity frequency
    delta = 1.0 * 2 * np.pi   # atom frequency
    g2 = 0.1
    g1 = 0.0
    psi0 = basis(2, 0)        # initial state
    tlist = np.linspace(0, 5, 200)

    sx, sy, sz = _qubit_integrate(tlist, psi0, epsilon, delta, g1, g2, "me")

    sx_analytic = np.zeros(np.shape(tlist))
    sy_analytic = -np.sin(2 * np.pi * tlist) * np.exp(-tlist * g2)
    sz_analytic = np.cos(2 * np.pi * tlist) * np.exp(-tlist * g2)

    assert_equal(max(abs(sx - sx_analytic)) < 0.05, True)
    assert_equal(max(abs(sy - sy_analytic)) < 0.05, True)
    assert_equal(max(abs(sz - sz_analytic)) < 0.05, True)


def test_MESolverCase2():
    """
    Test mesolve qubit, no dissipation
    """

    epsilon = 0.0 * 2 * np.pi   # cavity frequency
    delta = 1.0 * 2 * np.pi   # atom frequency
    g2 = 0.0
    g1 = 0.0
    psi0 = basis(2, 0)        # initial state
    tlist = np.linspace(0, 5, 200)

    sx, sy, sz = _qubit_integrate(tlist, psi0, epsilon, delta, g1, g2, "me")

    sx_analytic = np.zeros(np.shape(tlist))
    sy_analytic = -np.sin(2 * np.pi * tlist) * np.exp(-tlist * g2)
    sz_analytic = np.cos(2 * np.pi * tlist) * np.exp(-tlist * g2)

    assert_equal(max(abs(sx - sx_analytic)) < 0.05, True)
    assert_equal(max(abs(sy - sy_analytic)) < 0.05, True)
    assert_equal(max(abs(sz - sz_analytic)) < 0.05, True)


def test_ESSolverCase1():
    """
    Test essolve qubit, with dissipation
    """
    epsilon = 0.0 * 2 * np.pi      # cavity frequency
    delta = 1.0 * 2 * np.pi        # atom frequency
    g2 = 0.1
    g1 = 0.0
    psi0 = basis(2, 0)          # initial state
    tlist = np.linspace(0, 5, 200)

    with pytest.deprecated_call():
        sx, sy, sz = _qubit_integrate(tlist, psi0, epsilon, delta, g1, g2, "es")

    sx_analytic = np.zeros(np.shape(tlist))
    sy_analytic = -np.sin(2 * np.pi * tlist) * np.exp(-tlist * g2)
    sz_analytic = np.cos(2 * np.pi * tlist) * np.exp(-tlist * g2)

    assert_equal(max(abs(sx - sx_analytic)) < 0.05, True)
    assert_equal(max(abs(sy - sy_analytic)) < 0.05, True)
    assert_equal(max(abs(sz - sz_analytic)) < 0.05, True)


def test_MCSolverCase1():
    """
    Test mcsolve qubit, with dissipation
    """

    epsilon = 0.0 * 2 * np.pi      # cavity frequency
    delta = 1.0 * 2 * np.pi        # atom frequency
    g2 = 0.1
    g1 = 0.0
    psi0 = basis(2, 0)          # initial state
    tlist = np.linspace(0, 5, 200)

    sx, sy, sz = _qubit_integrate(tlist, psi0, epsilon, delta, g1, g2, "mc")

    sx_analytic = np.zeros(np.shape(tlist))
    sy_analytic = -np.sin(2 * np.pi * tlist) * np.exp(-tlist * g2)
    sz_analytic = np.cos(2 * np.pi * tlist) * np.exp(-tlist * g2)

    assert_equal(max(abs(sx - sx_analytic)) < 0.25, True)
    assert_equal(max(abs(sy - sy_analytic)) < 0.25, True)
    assert_equal(max(abs(sz - sz_analytic)) < 0.25, True)


def test_MCSolverCase2():
    """
    Test mcsolve qubit, no dissipation
    """

    epsilon = 0.0 * 2 * np.pi      # cavity frequency
    delta = 1.0 * 2 * np.pi        # atom frequency
    g2 = 0.0
    g1 = 0.0
    psi0 = basis(2, 0)          # initial state
    tlist = np.linspace(0, 5, 200)

    sx, sy, sz = _qubit_integrate(tlist, psi0, epsilon, delta, g1, g2, "mc")

    sx_analytic = np.zeros(np.shape(tlist))
    sy_analytic = -np.sin(2 * np.pi * tlist) * np.exp(-tlist * g2)
    sz_analytic = np.cos(2 * np.pi * tlist) * np.exp(-tlist * g2)

    assert_equal(max(abs(sx - sx_analytic)) < 0.25, True)
    assert_equal(max(abs(sy - sy_analytic)) < 0.25, True)
    assert_equal(max(abs(sz - sz_analytic)) < 0.25, True)


if __name__ == "__main__":
    run_module_suite()



# -*- coding: utf-8 -*-
# This file is part of QuTiP: Quantum Toolbox in Python.
#
#    Copyright (c) 2014 and later, Alexander J G Pitchford
#    All rights reserved.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions are
#    met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of the QuTiP: Quantum Toolbox in Python nor the names
#       of its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
###############################################################################

# @author: Alexander Pitchford
# @email1: agp1@aber.ac.uk
# @email2: alex.pitchford@gmail.com
# @organization: Aberystwyth University
# @supervisor: Daniel Burgarth

"""
Statistics for the optimisation
Note that some of the stats here are redundant copies from the optimiser
used here for calculations
"""
import numpy as np
import datetime


class Stats(object):
    """
    Base class for all optimisation statistics
    Used for configurations where all timeslots are updated each iteration
    e.g. exact gradients
    Note that all times are generated using timeit.default_timer() and are
    in seconds

    Attributes
    ----------
    dyn_gen_name : string
        Text used in some report functions.
        Makes sense to set it to 'Hamiltonian' when using unitary dynamics
        Default is simply 'dynamics generator'

    num_iter : integer
        Number of iterations of the optimisation algorithm

    wall_time_optim_start : float
        Start time for the optimisation

    wall_time_optim_end : float
        End time for the optimisation

    wall_time_optim : float
        Time elasped during the optimisation

    wall_time_dyn_gen_compute : float
        Total wall (elasped) time computing combined dynamics generator
        (for example combining drift and control Hamiltonians)

    wall_time_prop_compute : float
        Total wall (elasped) time computing propagators, that is the
        time evolution from one timeslot to the next
        Includes calculating the propagator gradient for exact gradients

    wall_time_fwd_prop_compute : float
        Total wall (elasped) time computing combined forward propagation,
        that is the time evolution from the start to a specific timeslot.
        Excludes calculating the propagators themselves

    wall_time_onwd_prop_compute : float
        Total wall (elasped) time computing combined onward propagation,
        that is the time evolution from a specific timeslot to the end time.
        Excludes calculating the propagators themselves

    wall_time_gradient_compute : float
        Total wall (elasped) time computing the fidelity error gradient.
        Excludes calculating the propagator gradients (in exact gradient
        methods)

    num_fidelity_func_calls : integer
        Number of calls to fidelity function by the optimisation algorithm

    num_grad_func_calls : integer
        Number of calls to gradient function by the optimisation algorithm

    num_tslot_recompute : integer
        Number of time the timeslot evolution is recomputed
        (It is only computed if any amplitudes changed since the last call)

    num_fidelity_computes : integer
        Number of time the fidelity is computed
        (It is only computed if any amplitudes changed since the last call)

    num_grad_computes : integer
        Number of time the gradient is computed
        (It is only computed if any amplitudes changed since the last call)

    num_ctrl_amp_updates : integer
        Number of times the control amplitudes are updated

    mean_num_ctrl_amp_updates_per_iter : float
        Mean number of control amplitude updates per iteration

    num_timeslot_changes : integer
        Number of times the amplitudes of a any control in a timeslot changes

    mean_num_timeslot_changes_per_update : float
        Mean average number of timeslot amplitudes that are changed per update

    num_ctrl_amp_changes : integer
        Number of times individual control amplitudes that are changed

    mean_num_ctrl_amp_changes_per_update : float
        Mean average number of control amplitudes that are changed per update
    """
    def __init__(self):
        self.reset()

    def reset(self):
        self.dyn_gen_name = "dynamics generator"
        self.clear()

    def clear(self):
        self.num_iter = 0
        # Duration attributes
        self.wall_time_optim_start = 0.0
        self.wall_time_optim_end = 0.0
        self.wall_time_optim = 0.0
        self.wall_time_dyn_gen_compute = 0.0
        self.wall_time_prop_compute = 0.0
        self.wall_time_fwd_prop_compute = 0.0
        self.wall_time_onwd_prop_compute = 0.0
        self.wall_time_gradient_compute = 0.0
        # Fidelity and gradient function calls and computes
        self.num_fidelity_func_calls = 0
        self.num_grad_func_calls = 0
        self.num_tslot_recompute = 0
        self.num_fidelity_computes = 0
        self.num_grad_computes = 0
        # Control amplitudes
        self.num_ctrl_amp_updates = 0
        self.mean_num_ctrl_amp_updates_per_iter = 0.0
        self.num_timeslot_changes = 0
        self.mean_num_timeslot_changes_per_update = 0.0
        self.num_ctrl_amp_changes = 0
        self.mean_num_ctrl_amp_changes_per_update = 0.0

    def calculate(self):
        """
        Perform the calculations (e.g. averages) that are required on the stats
        Should be called before calling report
        """
        # If the optimation is still running then the optimisation
        # time is the time so far
        if self.wall_time_optim_end > 0.0:
            self.wall_time_optim = \
                self.wall_time_optim_end - self.wall_time_optim_start

        try:
            self.mean_num_ctrl_amp_updates_per_iter = \
                self.num_ctrl_amp_updates / float(self.num_iter)
        except:
            self.mean_num_ctrl_amp_updates_per_iter = np.NaN

        try:
            self.mean_num_timeslot_changes_per_update = \
                self.num_timeslot_changes / float(self.num_ctrl_amp_updates)
        except:
            self.mean_num_timeslot_changes_per_update = np.NaN

        try:
            self.mean_num_ctrl_amp_changes_per_update = \
                self.num_ctrl_amp_changes / float(self.num_ctrl_amp_updates)
        except:
            self.mean_num_ctrl_amp_changes_per_update = np.NaN

    def _format_datetime(self, t, tot=0.0):
        dtStr = str(datetime.timedelta(seconds=t))
        if tot > 0:
            percent = 100*t/tot
            dtStr += " ({:03.2f}%)".format(percent)
        return dtStr

    def report(self):
        """
        Print a report of the stats to the console
        """
        print("\n------------------------------------"
              "\n---- Control optimisation stats ----")
        self.report_timings()
        self.report_func_calls()
        self.report_amp_updates()
        print("------------------------------------")

    def report_timings(self):
        print("**** Timings (HH:MM:SS.US) ****")
        tot = self.wall_time_optim
        print("Total wall time elapsed during optimisation: " +
              self._format_datetime(tot))
        print("Wall time computing Hamiltonians: " +
              self._format_datetime(self.wall_time_dyn_gen_compute, tot))
        print("Wall time computing propagators: " +
              self._format_datetime(self.wall_time_prop_compute, tot))
        print("Wall time computing forward propagation: " +
              self._format_datetime(self.wall_time_fwd_prop_compute, tot))
        print("Wall time computing onward propagation: " +
              self._format_datetime(self.wall_time_onwd_prop_compute, tot))
        print("Wall time computing gradient: " +
              self._format_datetime(self.wall_time_gradient_compute, tot))
        print("")

    def report_func_calls(self):
        print("**** Iterations and function calls ****")
        print("Number of iterations: {}".format(self.num_iter))
        print("Number of fidelity function calls: "
              "{}".format(self.num_fidelity_func_calls))
        print("Number of times fidelity is computed: "
              "{}".format(self.num_fidelity_computes))
        print("Number of gradient function calls: "
              "{}".format(self.num_grad_func_calls))
        print("Number of times gradients are computed: "
              "{}".format(self.num_grad_computes))
        print("Number of times timeslot evolution is recomputed: "
              "{}".format(self.num_tslot_recompute))
        print("")

    def report_amp_updates(self):
        print("**** Control amplitudes ****")
        print("Number of control amplitude updates: "
              "{}".format(self.num_ctrl_amp_updates))
        print("Mean number of updates per iteration: "
              "{}".format(self.mean_num_ctrl_amp_updates_per_iter))
        print("Number of timeslot values changed: "
              "{}".format(self.num_timeslot_changes))
        print("Mean number of timeslot changes per update: "
              "{}".format(self.mean_num_timeslot_changes_per_update))
        print("Number of amplitude values changed: "
              "{}".format(self.num_ctrl_amp_changes))
        print("Mean number of amplitude changes per update: "
              "{}".format(self.mean_num_ctrl_amp_changes_per_update))


class StatsDynTsUpdate(Stats):
    """
    Optimisation stats class for configurations where all timeslots are not
    necessarily updated at each iteration. In this case it may be interesting
    to know how many Hamiltions etc are computed each ctrl amplitude update

    Attributes
    ----------
    num_dyn_gen_computes : integer
        Total number of dynamics generator (Hamiltonian) computations,
        that is combining drift and control dynamics to calculate the
        combined dynamics generator for the timeslot

    mean_num_dyn_gen_computes_per_update : float
        # Mean average number of dynamics generator computations per update

    mean_wall_time_dyn_gen_compute : float
        # Mean average time to compute a timeslot dynamics generator

    num_prop_computes : integer
        Total number of propagator (and propagator gradient for exact
        gradient types) computations

    mean_num_prop_computes_per_update : float
        Mean average number of propagator computations per update

    mean_wall_time_prop_compute : float
        Mean average time to compute a propagator (and its gradient)

    num_fwd_prop_step_computes : integer
        Total number of steps (matrix product) computing forward propagation

    mean_num_fwd_prop_step_computes_per_update : float
        Mean average number of steps computing forward propagation

    mean_wall_time_fwd_prop_compute : float
        Mean average time to compute forward propagation

    num_onwd_prop_step_computes : integer
        Total number of steps (matrix product) computing onward propagation

    mean_num_onwd_prop_step_computes_per_update : float
        Mean average number of steps computing onward propagation

    mean_wall_time_onwd_prop_compute
        Mean average time to compute onward propagation
    """

    def __init__(self):
        self.reset()

    def reset(self):
        Stats.reset(self)
        # Dynamics generators (Hamiltonians)
        self.num_dyn_gen_computes = 0
        self.mean_num_dyn_gen_computes_per_update = 0.0
        self.mean_wall_time_dyn_gen_compute = 0.0
        # **** Propagators *****
        self.num_prop_computes = 0
        self.mean_num_prop_computes_per_update = 0.0
        self.mean_wall_time_prop_compute = 0.0
        # **** Forward propagation ****
        self.num_fwd_prop_step_computes = 0
        self.mean_num_fwd_prop_step_computes_per_update = 0.0
        self.mean_wall_time_fwd_prop_compute = 0.0
        # **** onward propagation ****
        self.num_onwd_prop_step_computes = 0
        self.mean_num_onwd_prop_step_computes_per_update = 0.0
        self.mean_wall_time_onwd_prop_compute = 0.0

    def calculate(self):
        Stats.calculate(self)
        self.mean_num_dyn_gen_computes_per_update = \
            self.num_dyn_gen_computes / float(self.num_ctrl_amp_updates)

        self.mean_wall_time_dyn_gen_compute = \
            (self.wall_time_dyn_gen_compute /
             float(self.num_dyn_gen_computes))

        self.mean_num_prop_computes_per_update = \
            self.num_prop_computes / float(self.num_ctrl_amp_updates)

        self.mean_wall_time_prop_compute = \
            self.wall_time_prop_compute / float(self.num_prop_computes)

        self.mean_num_fwd_prop_step_computes_per_update = \
            (self.num_fwd_prop_step_computes /
             float(self.num_ctrl_amp_updates))

        self.mean_wall_time_fwd_prop_compute = \
            (self.wall_time_fwd_prop_compute /
             float(self.num_fwd_prop_step_computes))

        self.mean_num_onwd_prop_step_computes_per_update = \
            (self.num_onwd_prop_step_computes /
             float(self.num_ctrl_amp_updates))

        self.mean_wall_time_onwd_prop_compute = \
            (self.wall_time_onwd_prop_compute /
             float(self.num_onwd_prop_step_computes))

    def report(self):
        """
        Print a report of the stats to the console
        """

        print("\n------------------------------------"
              "\n---- Control optimisation stats ----")
        self.report_timings()
        self.report_func_calls()
        self.report_amp_updates()
        self.report_dyn_gen_comps()
        self.report_fwd_prop()
        self.report_onwd_prop()
        print("------------------------------------")

    def report_dyn_gen_comps(self):
        print("**** {} Computations ****".format(self.dyn_gen_name))
        print("Total number of {} computations: "
              "{}".format(self.dyn_gen_name, self.num_dyn_gen_computes))
        print("Mean number of {} computations per update: "
              "{}".format(self.dyn_gen_name,
                          self.mean_num_dyn_gen_computes_per_update))
        print("Mean wall time to compute {}s: "
              "{} s".format(self.dyn_gen_name,
                            self.mean_wall_time_dyn_gen_compute))
        print("**** Propagator Computations ****")
        print("Total number of propagator computations: "
              "{}".format(self.num_prop_computes))
        print("Mean number of propagator computations per update: "
              "{}".format(self.mean_num_prop_computes_per_update))
        print("Mean wall time to compute propagator "
              "{} s".format(self.mean_wall_time_prop_compute))

    def report_fwd_prop(self):
        print("**** Forward Propagation ****")
        print("Total number of forward propagation step computations: "
              "{}".format(self.num_fwd_prop_step_computes))
        print("Mean number of forward propagation step computations"
              " per update: "
              "{}".format(self.mean_num_fwd_prop_step_computes_per_update))
        print("Mean wall time to compute forward propagation "
              "{} s".format(self.mean_wall_time_fwd_prop_compute))

    def report_onwd_prop(self):
        print("**** Onward Propagation ****")
        print("Total number of onward propagation step computations: "
              "{}".format(self.num_onwd_prop_step_computes))
        print("Mean number of onward propagation step computations"
              " per update: "
              "{}".format(self.mean_num_onwd_prop_step_computes_per_update))
        print("Mean wall time to compute onward propagation "
              "{} s".format(self.mean_wall_time_onwd_prop_compute))



# This file is part of QuTiP: Quantum Toolbox in Python.
#
#    Copyright (c) 2011 and later, Paul D. Nation and Robert J. Johansson.
#    All rights reserved.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions are
#    met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of the QuTiP: Quantum Toolbox in Python nor the names
#       of its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
###############################################################################

"""
This module provides classes and functions for working with spatial
distributions, such as Wigner distributions, etc.

.. note::

    Experimental.

"""

__all__ = ['Distribution', 'WignerDistribution', 'QDistribution',
           'TwoModeQuadratureCorrelation',
           'HarmonicOscillatorWaveFunction',
           'HarmonicOscillatorProbabilityFunction']

import numpy as np
from numpy import pi, exp, sqrt

from scipy.special import hermite, factorial

from qutip.qobj import isket
from qutip.wigner import wigner, qfunc
from qutip.states import ket2dm, state_number_index

try:
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
except:
    pass


class Distribution:
    """A class for representation spatial distribution functions.

    The Distribution class can be used to prepresent spatial distribution
    functions of arbitray dimension (although only 1D and 2D distributions
    are used so far).

    It is indented as a base class for specific distribution function, and
    provide implementation of basic functions that are shared among all
    Distribution functions, such as visualization, calculating marginal
    distributions, etc.

    Parameters
    ----------
    data : array_like
        Data for the distribution. The dimensions must match the lengths of
        the coordinate arrays in xvecs.
    xvecs : list
        List of arrays that spans the space for each coordinate.
    xlabels : list
        List of labels for each coordinate.

    """

    def __init__(self, data=None, xvecs=[], xlabels=[]):
        self.data = data
        self.xvecs = xvecs
        self.xlabels = xlabels

    def visualize(self, fig=None, ax=None, figsize=(8, 6),
                  colorbar=True, cmap=None, style="colormap",
                  show_xlabel=True, show_ylabel=True):
        """
        Visualize the data of the distribution in 1D or 2D, depending
        on the dimensionality of the underlaying distribution.

        Parameters:

        fig : matplotlib Figure instance
            If given, use this figure instance for the visualization,

        ax : matplotlib Axes instance
            If given, render the visualization using this axis instance.

        figsize : tuple
            Size of the new Figure instance, if one needs to be created.

        colorbar: Bool
            Whether or not the colorbar (in 2D visualization) should be used.

        cmap: matplotlib colormap instance
            If given, use this colormap for 2D visualizations.

        style : string
            Type of visualization: 'colormap' (default) or 'surface'.

        Returns
        -------

        fig, ax : tuple
            A tuple of matplotlib figure and axes instances.

        """
        n = len(self.xvecs)
        if n == 2:
            if style == "colormap":
                return self.visualize_2d_colormap(fig=fig, ax=ax,
                                                  figsize=figsize,
                                                  colorbar=colorbar,
                                                  cmap=cmap,
                                                  show_xlabel=show_xlabel,
                                                  show_ylabel=show_ylabel)
            else:
                return self.visualize_2d_surface(fig=fig, ax=ax,
                                                 figsize=figsize,
                                                 colorbar=colorbar,
                                                 cmap=cmap,
                                                 show_xlabel=show_xlabel,
                                                 show_ylabel=show_ylabel)

        elif n == 1:
            return self.visualize_1d(fig=fig, ax=ax, figsize=figsize,
                                     show_xlabel=show_xlabel,
                                     show_ylabel=show_ylabel)
        else:
            raise NotImplementedError("Distribution visualization in " +
                                      "%d dimensions is not implemented." % n)

    def visualize_2d_colormap(self, fig=None, ax=None, figsize=(8, 6),
                              colorbar=True, cmap=None,
                              show_xlabel=True, show_ylabel=True):

        if not fig and not ax:
            fig, ax = plt.subplots(1, 1, figsize=figsize)

        if cmap is None:
            cmap = mpl.cm.get_cmap('RdBu')

        lim = abs(self.data).max()

        cf = ax.contourf(self.xvecs[0], self.xvecs[1], self.data, 100,
                         norm=mpl.colors.Normalize(-lim, lim),
                         cmap=cmap)

        if show_xlabel:
            ax.set_xlabel(self.xlabels[0], fontsize=12)
        if show_ylabel:
            ax.set_ylabel(self.xlabels[1], fontsize=12)

        if colorbar:
            cb = fig.colorbar(cf, ax=ax)

        return fig, ax

    def visualize_2d_surface(self, fig=None, ax=None, figsize=(8, 6),
                             colorbar=True, cmap=None,
                             show_xlabel=True, show_ylabel=True):

        if not fig and not ax:
            fig = plt.figure(figsize=figsize)
            ax = Axes3D(fig, azim=-62, elev=25)

        if cmap is None:
            cmap = mpl.cm.get_cmap('RdBu')

        lim = abs(self.data).max()

        X, Y = np.meshgrid(self.xvecs[0], self.xvecs[1])
        s = ax.plot_surface(X, Y, self.data,
                            norm=mpl.colors.Normalize(-lim, lim),
                            rstride=5, cstride=5, cmap=cmap, lw=0.1)

        if show_xlabel:
            ax.set_xlabel(self.xlabels[0], fontsize=12)
        if show_ylabel:
            ax.set_ylabel(self.xlabels[1], fontsize=12)

        if colorbar:
            cb = fig.colorbar(s, ax=ax, shrink=0.5)

        return fig, ax

    def visualize_1d(self, fig=None, ax=None, figsize=(8, 6),
                     show_xlabel=True, show_ylabel=True):

        if not fig and not ax:
            fig, ax = plt.subplots(1, 1, figsize=figsize)

        p = ax.plot(self.xvecs[0], self.data)

        if show_xlabel:
            ax.set_xlabel(self.xlabels[0], fontsize=12)
        if show_ylabel:
            ax.set_ylabel("Marginal distribution", fontsize=12)

        return fig, ax

    def marginal(self, dim=0):
        """
        Calculate the marginal distribution function along the dimension
        `dim`. Return a new Distribution instance describing this reduced-
        dimensionality distribution.

        Parameters
        ----------
        dim : int
            The dimension (coordinate index) along which to obtain the
            marginal distribution.

        Returns
        -------

        d : Distributions
            A new instances of Distribution that describes the marginal
            distribution.

        """
        return Distribution(data=self.data.mean(axis=dim),
                            xvecs=[self.xvecs[dim]],
                            xlabels=[self.xlabels[dim]])

    def project(self, dim=0):
        """
        Calculate the projection (max value) distribution function along the
        dimension `dim`. Return a new Distribution instance describing this
        reduced-dimensionality distribution.

        Parameters
        ----------
        dim : int
            The dimension (coordinate index) along which to obtain the
            projected distribution.

        Returns
        -------
        d : Distributions
            A new instances of Distribution that describes the projection.

        """
        return Distribution(data=self.data.max(axis=dim),
                            xvecs=[self.xvecs[dim]],
                            xlabels=[self.xlabels[dim]])


class WignerDistribution(Distribution):

    def __init__(self, rho=None, extent=[[-5, 5], [-5, 5]], steps=250):

        self.xvecs = [np.linspace(extent[0][0], extent[0][1], steps),
                      np.linspace(extent[1][0], extent[1][1], steps)]

        self.xlabels = [r'$\rm{Re}(\alpha)$', r'$\rm{Im}(\alpha)$']

        if rho:
            self.update(rho)

    def update(self, rho):

        self.data = wigner(rho, self.xvecs[0], self.xvecs[1])


class QDistribution(Distribution):

    def __init__(self, rho=None, extent=[[-5, 5], [-5, 5]], steps=250):

        self.xvecs = [np.linspace(extent[0][0], extent[0][1], steps),
                      np.linspace(extent[1][0], extent[1][1], steps)]

        self.xlabels = [r'$\rm{Re}(\alpha)$', r'$\rm{Im}(\alpha)$']

        if rho:
            self.update(rho)

    def update(self, rho):

        self.data = qfunc(rho, self.xvecs[0], self.xvecs[1])


class TwoModeQuadratureCorrelation(Distribution):

    def __init__(self, state=None, theta1=0.0, theta2=0.0,
                 extent=[[-5, 5], [-5, 5]], steps=250):

        self.xvecs = [np.linspace(extent[0][0], extent[0][1], steps),
                      np.linspace(extent[1][0], extent[1][1], steps)]

        self.xlabels = [r'$X_1(\theta_1)$', r'$X_2(\theta_2)$']

        self.theta1 = theta1
        self.theta2 = theta2

        self.update(state)

    def update(self, state):
        """
        calculate probability distribution for quadrature measurement
        outcomes given a two-mode wavefunction or density matrix
        """
        if isket(state):
            self.update_psi(state)
        else:
            self.update_rho(state)

    def update_psi(self, psi):
        """
        calculate probability distribution for quadrature measurement
        outcomes given a two-mode wavefunction
        """

        X1, X2 = np.meshgrid(self.xvecs[0], self.xvecs[1])

        p = np.zeros((len(self.xvecs[0]), len(self.xvecs[1])), dtype=complex)
        N = psi.dims[0][0]

        for n1 in range(N):
            kn1 = exp(-1j * self.theta1 * n1) / \
                sqrt(sqrt(pi) * 2 ** n1 * factorial(n1)) * \
                exp(-X1 ** 2 / 2.0) * np.polyval(hermite(n1), X1)

            for n2 in range(N):
                kn2 = exp(-1j * self.theta2 * n2) / \
                    sqrt(sqrt(pi) * 2 ** n2 * factorial(n2)) * \
                    exp(-X2 ** 2 / 2.0) * np.polyval(hermite(n2), X2)
                i = state_number_index([N, N], [n1, n2])
                p += kn1 * kn2 * psi.data[i, 0]

        self.data = abs(p) ** 2

    def update_rho(self, rho):
        """
        calculate probability distribution for quadrature measurement
        outcomes given a two-mode density matrix
        """

        X1, X2 = np.meshgrid(self.xvecs[0], self.xvecs[1])

        p = np.zeros((len(self.xvecs[0]), len(self.xvecs[1])), dtype=complex)
        N = rho.dims[0][0]

        M1 = np.zeros(
            (N, N, len(self.xvecs[0]), len(self.xvecs[1])), dtype=complex)
        M2 = np.zeros(
            (N, N, len(self.xvecs[0]), len(self.xvecs[1])), dtype=complex)

        for m in range(N):
            for n in range(N):
                M1[m, n] = exp(-1j * self.theta1 * (m - n)) / \
                    sqrt(pi * 2 ** (m + n) * factorial(n) * factorial(m)) * \
                    exp(-X1 ** 2) * np.polyval(
                        hermite(m), X1) * np.polyval(hermite(n), X1)
                M2[m, n] = exp(-1j * self.theta2 * (m - n)) / \
                    sqrt(pi * 2 ** (m + n) * factorial(n) * factorial(m)) * \
                    exp(-X2 ** 2) * np.polyval(
                        hermite(m), X2) * np.polyval(hermite(n), X2)

        for n1 in range(N):
            for n2 in range(N):
                i = state_number_index([N, N], [n1, n2])
                for p1 in range(N):
                    for p2 in range(N):
                        j = state_number_index([N, N], [p1, p2])
                        p += M1[n1, p1] * M2[n2, p2] * rho.data[i, j]

        self.data = p


class HarmonicOscillatorWaveFunction(Distribution):

    def __init__(self, psi=None, omega=1.0, extent=[-5, 5], steps=250):

        self.xvecs = [np.linspace(extent[0], extent[1], steps)]
        self.xlabels = [r'$x$']
        self.omega = omega

        if psi:
            self.update(psi)

    def update(self, psi):
        """
        Calculate the wavefunction for the given state of an harmonic
        oscillator
        """

        self.data = np.zeros(len(self.xvecs[0]), dtype=complex)
        N = psi.shape[0]

        for n in range(N):
            k = pow(self.omega / pi, 0.25) / \
                sqrt(2 ** n * factorial(n)) * \
                exp(-self.xvecs[0] ** 2 / 2.0) * \
                np.polyval(hermite(n), self.xvecs[0])

            self.data += k * psi.data[n, 0]


class HarmonicOscillatorProbabilityFunction(Distribution):

    def __init__(self, rho=None, omega=1.0, extent=[-5, 5], steps=250):

        self.xvecs = [np.linspace(extent[0], extent[1], steps)]
        self.xlabels = [r'$x$']
        self.omega = omega

        if rho:
            self.update(rho)

    def update(self, rho):
        """
        Calculate the probability function for the given state of an harmonic
        oscillator (as density matrix)
        """

        if isket(rho):
            rho = ket2dm(rho)

        self.data = np.zeros(len(self.xvecs[0]), dtype=complex)
        M, N = rho.shape

        for m in range(M):
            k_m = pow(self.omega / pi, 0.25) / \
                sqrt(2 ** m * factorial(m)) * \
                exp(-self.xvecs[0] ** 2 / 2.0) * \
                np.polyval(hermite(m), self.xvecs[0])

            for n in range(N):
                k_n = pow(self.omega / pi, 0.25) / \
                    sqrt(2 ** n * factorial(n)) * \
                    exp(-self.xvecs[0] ** 2 / 2.0) * \
                    np.polyval(hermite(n), self.xvecs[0])

                self.data += np.conjugate(k_n) * k_m * rho.data[m, n]
